# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

resources:
- name: {{ env["deployment"] }}-instance
  type: gcp-types/redis-v1:projects.locations.instances
  properties:
    parent: projects/{{ env["project"] }}/locations/{{ properties["region"] }}
    instanceId: {{ env["deployment"] }}-instance
    authorizedNetwork: projects/{{ env["project"] }}/global/networks/default
    memorySizeGb: {{ properties["memorySizeGb"] }}
    tier: BASIC
    {% if properties["displayName"] %}
    displayName: {{ properties["displayName"] }}
    {% endif %}
